
from asyncio import Future, get_event_loop
from rx.internal import extensionmethod
from rx.core import Observable


async def __aiter__(self):
    source = self

    class AIterator:
        def __init__(self):
            self.notifications = []
            self.future = Future()

            self.disposable = source.materialize().subscribe(self.on_next)
            # self.disposed = False

        def __aiter__(self):
            return self

        def dispose(self):
            # self.future.cancel()
            # self.disposed = True
            # self.future.set_exception(StopAsyncIteration)
            self.disposable.dispose()

        def feeder(self):
            if not self.notifications or self.future.done():
                return

            notification = self.notifications.pop(0)
            kind = notification.kind
            if kind == 'N':
                self.future.set_result(notification.value)
            if kind == 'E':
                self.future.set_exception(notification.exception)
            if kind == 'C':
                self.future.set_exception(StopAsyncIteration)

        def on_next(self, notification):
            self.notifications.append(notification)
            self.feeder()

        async def __anext__(self):
            # if self.disposed:
            #     raise StopAsyncIteration
            self.feeder()

            value = await self.future
            self.future = Future()
            return value

    return AIterator()


# def __aiter__(self, sentinel=None):
#     loop = get_event_loop()
#     future = [Future()]
#     notifications = []

#     def feeder():
#         if not len(notifications) or future[0].done():
#             return
#         notification = notifications.pop(0)
#         if notification.kind == "E":
#             future[0].set_exception(notification.exception)
#         elif notification.kind == "C":
#             future[0].set_exception(StopIteration(sentinel))
#         else:
#             future[0].set_result(notification.value)

#     def on_next(value):
#         """Takes on_next values and appends them to the notification queue"""
#         notifications.append(value)
#         loop.call_soon(feeder)

#     self.materialize().subscribe(on_next)

#     @asyncio.coroutine
#     def gen():
#         """Generator producing futures"""
#         loop.call_soon(feeder)
#         future[0] = Future()
#         return future[0]

#     return gen


def setup_observable_extension():
    extensionmethod(Observable)(__aiter__)



import json

from channels.generic.websocket import JsonWebsocketConsumer
from django.conf import settings
from graphene_django.settings import graphene_settings
from graphql.execution.executors.sync import SyncExecutor
from rx import Observable, Observer

from .base import (BaseConnectionContext, BaseSubscriptionServer,
                   ConnectionClosedException)
from .constants import GQL_CONNECTION_ACK, GQL_CONNECTION_ERROR


class DjangoChannelConnectionContext(BaseConnectionContext):

    def __init__(self, message, request_context=None):
        self.message = message
        self.operations = {}
        self.request_context = request_context

    def send(self, data):
        self.message.reply_channel.send(data)

    def close(self, reason):
        data = {
            'close': True,
            'text': reason
        }
        self.message.reply_channel.send(data)


class DjangoChannelSubscriptionServer(BaseSubscriptionServer):

    def get_graphql_params(self, *args, **kwargs):
        params = super(DjangoChannelSubscriptionServer,
                       self).get_graphql_params(*args, **kwargs)
        return dict(params, executor=SyncExecutor())

    def handle(self, message, connection_context):
        self.on_message(connection_context, message)

    def send_message(self, connection_context, op_id=None, op_type=None, payload=None):
        message = {}
        if op_id is not None:
            message['id'] = op_id
        if op_type is not None:
            message['type'] = op_type
        if payload is not None:
            message['payload'] = payload

        assert message, "You need to send at least one thing"
        return connection_context.send({'text': json.dumps(message)})

    def on_open(self, connection_context):
        pass

    def on_connect(self, connection_context, payload):
        pass

    def on_connection_init(self, connection_context, op_id, payload):
        try:
            self.on_connect(connection_context, payload)
            self.send_message(connection_context, op_type=GQL_CONNECTION_ACK)

        except Exception as e:
            self.send_error(connection_context, op_id, e, GQL_CONNECTION_ERROR)
            connection_context.close(1011)

    def on_start(self, connection_context, op_id, params):
        try:
            execution_result = self.execute(
                connection_context.request_context, params)
            assert isinstance(
                execution_result, Observable), "A subscription must return an observable"
            execution_result.subscribe(SubscriptionObserver(
                connection_context,
                op_id,
                self.send_execution_result,
                self.send_error,
                self.on_close
            ))
        except Exception as e:
            self.send_error(connection_context, op_id, str(e))

    def on_close(self, connection_context):
        remove_operations = list(connection_context.operations.keys())
        for op_id in remove_operations:
            self.unsubscribe(connection_context, op_id)

    def on_stop(self, connection_context, op_id):
        self.unsubscribe(connection_context, op_id)


class GraphQLSubscriptionConsumer(JsonWebsocketConsumer):
    http_user_and_session = True
    strict_ordering = True

    def connect(self, message, **kwargs):
        message.reply_channel.send({"accept": True})

    def receive(self, content, **kwargs):
        """
        Called when a message is received with either text or bytes
        filled out.
        """
        self.connection_context = DjangoChannelConnectionContext(self.message)
        self.subscription_server = DjangoChannelSubscriptionServer(
            graphene_settings.SCHEMA)
        self.subscription_server.on_open(self.connection_context)
        self.subscription_server.handle(content, self.connection_context)


class SubscriptionObserver(Observer):

    def __init__(self, connection_context, op_id, send_execution_result, send_error, on_close):
        self.connection_context = connection_context
        self.op_id = op_id
        self.send_execution_result = send_execution_result
        self.send_error = send_error
        self.on_close = on_close

    def on_next(self, value):
        self.send_execution_result(self.connection_context, self.op_id, value)

    def on_completed(self):
        self.on_close(self.connection_context)

    def on_error(self, error):
        self.send_error(self.connection_context, self.op_id, error)



GRAPHQL_WS = 'graphql-ws'
WS_PROTOCOL = GRAPHQL_WS

GQL_CONNECTION_INIT = 'connection_init'  # Client -> Server
GQL_CONNECTION_ACK = 'connection_ack'  # Server -> Client
GQL_CONNECTION_ERROR = 'connection_error'  # Server -> Client

# NOTE: This one here don't follow the standard due to connection optimization
GQL_CONNECTION_TERMINATE = 'connection_terminate'  # Client -> Server
GQL_CONNECTION_KEEP_ALIVE = 'ka'  # Server -> Client
GQL_START = 'start'  # Client -> Server
GQL_DATA = 'data'  # Server -> Client
GQL_ERROR = 'error'  # Server -> Client
GQL_COMPLETE = 'complete'  # Server -> Client
GQL_STOP = 'stop'  # Client -> Server



from __future__ import absolute_import

import json

from graphql import format_error, graphql
from graphql.execution.executors.sync import SyncExecutor
from rx import Observer, Observable
from .base import (
    ConnectionClosedException,
    BaseConnectionContext,
    BaseSubscriptionServer
)
from .constants import (
    GQL_CONNECTION_ACK,
    GQL_CONNECTION_ERROR
)


class GeventConnectionContext(BaseConnectionContext):

    def receive(self):
        msg = self.ws.receive()
        return msg

    def send(self, data):
        if self.closed:
            return
        self.ws.send(data)

    @property
    def closed(self):
        return self.ws.closed

    def close(self, code):
        self.ws.close(code)


class GeventSubscriptionServer(BaseSubscriptionServer):

    def get_graphql_params(self, *args, **kwargs):
        params = super(GeventSubscriptionServer,
                       self).get_graphql_params(*args, **kwargs)
        return dict(params, executor=SyncExecutor())

    def handle(self, ws, request_context=None):
        connection_context = GeventConnectionContext(ws, request_context)
        self.on_open(connection_context)
        while True:
            try:
                if connection_context.closed:
                    raise ConnectionClosedException()
                message = connection_context.receive()
            except ConnectionClosedException:
                self.on_close(connection_context)
                return
            self.on_message(connection_context, message)

    def on_open(self, connection_context):
        pass

    def on_connect(self, connection_context, payload):
        pass

    def on_close(self, connection_context):
        remove_operations = list(connection_context.operations.keys())
        for op_id in remove_operations:
            self.unsubscribe(connection_context, op_id)

    def on_connection_init(self, connection_context, op_id, payload):
        try:
            self.on_connect(connection_context, payload)
            self.send_message(connection_context, op_type=GQL_CONNECTION_ACK)

        except Exception as e:
            self.send_error(connection_context, op_id, e, GQL_CONNECTION_ERROR)
            connection_context.close(1011)

    def on_start(self, connection_context, op_id, params):
        try:
            execution_result = self.execute(
                connection_context.request_context, params)
            assert isinstance(
                execution_result, Observable), "A subscription must return an observable"
            execution_result.subscribe(SubscriptionObserver(
                connection_context,
                op_id,
                self.send_execution_result,
                self.send_error,
                self.on_close
            ))
        except Exception as e:
            self.send_error(connection_context, op_id, str(e))

    def on_stop(self, connection_context, op_id):
        self.unsubscribe(connection_context, op_id)


class SubscriptionObserver(Observer):

    def __init__(self, connection_context, op_id, send_execution_result, send_error, on_close):
        self.connection_context = connection_context
        self.op_id = op_id
        self.send_execution_result = send_execution_result
        self.send_error = send_error
        self.on_close = on_close

    def on_next(self, value):
        self.send_execution_result(self.connection_context, self.op_id, value)

    def on_completed(self):
        self.on_close(self.connection_context)

    def on_error(self, error):
        self.send_error(self.connection_context, self.op_id, error)



from distutils.core import setup

setup(
    name='graphql-ws-django',
    packages=['graphql-ws-django'],
    description='Websocket server for grapqh with django channels 2',
    version='0.3.1',
    author='Manviel',
    author_email='Mihaylo.Merezhko@kname.edu.ua',
    url='https://github.com/Manviel/graphql-ws',
    download_url='https://github.com/Manviel/graphql-ws',
    keywords=['graphql', 'channels'],
    install_requires=['grapqh-ws'],
)
